-- Lua optimizations (local references are faster than global look-ups)
local date = date
local format = format
local gsub = gsub
local strfind = strfind
local strsplit = strsplit
local strsub = strsub
local strupper = strupper
local time = time
local tinsert = tinsert
local tremove = tremove
local GetRealmName = GetRealmName
local GetNumSavedInstances = GetNumSavedInstances
local GetSavedInstanceInfo = GetSavedInstanceInfo
local SecondsToTime = SecondsToTime
local RequestRaidInfo = RequestRaidInfo
local UnitClass = UnitClass
local UnitName = UnitName

-- Use Blizzard's strings to save localization efforts
local NO_RAID_INSTANCES_SAVED = NO_RAID_INSTANCES_SAVED -- "You are not saved to any instances"
local FONTEND = FONT_COLOR_CODE_CLOSE
local GOLDFONT = NORMAL_FONT_COLOR_CODE
local YELLOWFONT = LIGHTYELLOW_FONT_COLOR_CODE
local WHITEFONT = HIGHLIGHT_FONT_COLOR_CODE
local GRAYFONT = GRAY_FONT_COLOR_CODE
local DUNGEON_DIFFICULTY1 = DUNGEON_DIFFICULTY1 -- 5 man
local DUNGEON_DIFFICULTY2 = DUNGEON_DIFFICULTY2 -- 5 man (Heroic)
local RAID_DIFFICULTY1 = RAID_DIFFICULTY1 -- "10 man"
local RAID_DIFFICULTY2 = RAID_DIFFICULTY2 -- "25 man"
local RAID_DIFFICULTY3 = RAID_DIFFICULTY3 -- "10 man (Heroic)"
local RAID_DIFFICULTY4 = RAID_DIFFICULTY4 -- "25 man (Heroic)"
local EXPANSION_NAME0 = EXPANSION_NAME0 -- "Classic"
local EXPANSION_NAME1 = EXPANSION_NAME1 -- "The Burning Crusade"
local EXPANSION_NAME2 = EXPANSION_NAME2 -- "Wrath of the Lich King"
local LFG_TYPE_DUNGEON = LFG_TYPE_DUNGEON -- "Dungeon"
local LFG_TYPE_RAID = LFG_TYPE_RAID -- "Raid"
local EMBLEM_SYMBOL = EMBLEM_SYMBOL -- "Icon"
local COLOR = COLOR -- "Color"
local DEFAULTS = DEFAULTS -- "Defaults"
local RESET_TO_DEFAULT = RESET_TO_DEFAULT -- "Reset to Default"
local DELETE = DELETE -- "DELETE"

local Indicators = {
	ICON_STAR = ICON_LIST[1] .. "16:16:0:0|t",
	ICON_CIRCLE = ICON_LIST[2] .. "16:16:0:0|t",
	ICON_DIAMOND = ICON_LIST[3] .. "16:16:0:0|t",
	ICON_TRIANGLE = ICON_LIST[4] .. "16:16:0:0|t",
	ICON_MOON = ICON_LIST[5] .. "16:16:0:0|t",
	ICON_SQUARE = ICON_LIST[6] .. "16:16:0:0|t",
	ICON_CROSS = ICON_LIST[7] .. "16:16:0:0|t",
	ICON_SKULL = ICON_LIST[8] .. "16:16:0:0|t",
	BLANK = "None",
}

local SavedInstances = LibStub("AceAddon-3.0"):NewAddon("SavedInstances", "AceEvent-3.0")
_G.SavedInstances = SavedInstances
local Config = LibStub("AceConfig-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("SavedInstances")

local LDB = LibStub("LibDataBroker-1.1", true)
local LDBIcon = LDB and LibStub("LibDBIcon-1.0", true)
local launcher

local QTip = LibStub("LibQTip-1.0")
local tooltip, indicatortip

local thisToon = UnitName("player") .. " - " .. GetRealmName()
local selectedInstance -- for re-ordering the instance list
local selectedCategory

local db
local defaultDB = {
	DBVersion = 7,
	Toons = { }, 	-- table key: "Toon - Realm"; value:
				-- Class: string
				-- Hide: boolean
	Indicators = {
		D1Indicator = "BLANK", -- indicator: ICON_*, BLANK
		D1Text = "N",
		D1Color = { 0, 0.6, 0, 1, }, -- dark green
		D2Indicator = "BLANK", -- indicator
		D2Text = "H",
		D2Color = { 0, 1, 0, 1, }, -- green
		R1Indicator = "BLANK", -- indicator: ICON_*, BLANK
		R1Text = "N",
		R1Color = { 0.6, 0.6, 0, 1, }, -- dark yellow
		R2Indicator = "BLANK", -- indicator
		R2Text = "H",
		R2Color = { 0.6, 0, 0, 1, }, -- dark red
		R3Indicator = "BLANK", -- indicator: ICON_*, BLANK
		R3Text = "N+",
		R3Color = { 1, 1, 0, 1, }, -- yellow
		R4Indicator = "BLANK", -- indicator
		R4Text = "H+",
		R4Color = { 1, 0, 0, 1, }, -- red
	},
	Tooltip = {
		FreshToons = false,
		AllRaids = true,
		Details = false,
		NewInstanceShow = false,
		ShowCategories = true,
		ShowSoloCategory = false,
		ShowHints = true,
	},
	Instances = { }, 	-- table key: "Instance name"; value:
					-- Show: boolean
					-- Order: integer
					-- Raid: boolean
					-- Expansion: integer
					-- table key: "Toon - Realm"; value:
						-- table key: "Difficulty"; value:
							-- ID: integer
							-- Expires: integer
	Categories = {
		D0 = { Name = EXPANSION_NAME0 .. ": " .. LFG_TYPE_DUNGEON, Order = 1 },
		R0 = { Name = EXPANSION_NAME0 .. ": " .. LFG_TYPE_RAID, Order = 2 },
		D1 = { Name = EXPANSION_NAME1 .. ": " .. LFG_TYPE_DUNGEON, Order = 3 },
		R1 = { Name = EXPANSION_NAME1 .. ": " .. LFG_TYPE_RAID, Order = 4 },
		D2 = { Name = EXPANSION_NAME2 .. ": " .. LFG_TYPE_DUNGEON, Order = 5 },
		R2 = { Name = EXPANSION_NAME2 .. ": " .. LFG_TYPE_RAID, Order = 6 },
	},
	MinimapIcon = { },
}

-- general helper functions below

local function ColorCodeOpen(color)
	return format("|c%02x%02x%02x%02x", math.floor(color[4] * 255), math.floor(color[1] * 255), math.floor(color[2] * 255), math.floor(color[3] * 255))
end

local function ClassColorise(class, targetstring)
	local RAID_CLASS_COLORS = CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS
	local color = {
		RAID_CLASS_COLORS[class].r,
		RAID_CLASS_COLORS[class].g,
		RAID_CLASS_COLORS[class].b,
		1,
	}
	return ColorCodeOpen(color) .. targetstring .. FONTEND
end

local function TableLen(table)
	local i = 0
	for _, _ in pairs(table) do
		i = i + 1
	end
	return i
end

-- local addon functions below

local function InstanceCategory(instance)
	if not instance then return nil end
	local instance = SavedInstances.db.Instances[instance]
	return ((instance.Raid and "R") or ((not instance.Raid) and "D")) .. instance.Expansion
end

local function InstancesInCategory(targetcategory)
	-- returns a table of the form { "instance1", "instance2", ... }
	if (not targetcategory) then return { } end
	local list = { }
	for instance, _ in pairs(SavedInstances.db.Instances) do
		if InstanceCategory(instance) == targetcategory then
			list[#list+1] = instance
		end
	end
	return list
end

local function CategorySize(category)
	if not category then return nil end
	local i = 0
	for instance, _ in pairs(SavedInstances.db.Instances) do
		if category == InstanceCategory(instance) then
			i = i + 1
		end
	end
	return i
end

local function OrderedInstances(category)
	-- returns a table of the form { "instance1", "instance2", ... }
	-- adjusts the Order value for each instance in the category
	if (not category) then return { } end
	local orderedlist = { }
	local instances = InstancesInCategory(category)
	while #orderedlist < CategorySize(category) and #instances > 0 do
		local nextinstance
		local nextinstanceindex
		for i, instance in ipairs(instances) do
			nextinstance = nextinstance or instance
			if SavedInstances.db.Instances[instance].Order <= SavedInstances.db.Instances[nextinstance].Order then
				nextinstance = instance
				nextinstanceindex = i
			end
		end
		orderedlist[#orderedlist+1] = nextinstance
		tremove(instances, nextinstanceindex)
	end
	for i, instance in ipairs(orderedlist) do
		SavedInstances.db.Instances[instance].Order = i
	end
	return orderedlist
end


local function OrderedCategories()
	-- returns a table of the form { "instance1", "instance2", ... }
	-- adjusts the Order value for each category
	local orderedlist = { }
	local categories = { }
	for category, v in pairs(SavedInstances.db.Categories) do
		categories[#categories+1] = category
	end
	while #orderedlist < TableLen(SavedInstances.db.Categories) and #categories > 0 do
		local nextcategory
		local nextinstanceindex
		for i, category in ipairs(categories) do
			nextcategory = nextcategory or category
			if SavedInstances.db.Categories[category].Order <= SavedInstances.db.Categories[nextcategory].Order then
				nextcategory = category
				nextcategoryindex = i
			end
		end
		orderedlist[#orderedlist+1] = nextcategory
		tremove(categories, nextcategoryindex)
	end
	for i, category in ipairs(orderedlist) do
		SavedInstances.db.Categories[category].Order = i
	end
	return orderedlist
end

local function ReorderInstance(category, from, to)
	-- modify the order of instances in their category given the starting and destination indices
	local instances = OrderedInstances(category)
	local movinginstance = instances[from]
	tremove(instances, from)
	tinsert(instances, to, movinginstance)
	for i, instance in ipairs(instances) do
		SavedInstances.db.Instances[instance].Order = i
--		print(instance .. tostring(SavedInstances.db.Instances[instance].Order))
	end
end

local function ReorderCategory(from, to)
	-- modify the order of categories given the starting and destination indices
	local categories = OrderedCategories()
	local movingcategory = categories[from]
	tremove(categories, from)
	tinsert(categories, to, movingcategory)
	for i, category in ipairs(categories) do
		SavedInstances.db.Categories[category].Order = i
	end
end

local function DifficultyString(instance, diff)
	local instance = SavedInstances.db.Instances[instance]
	local setting = ((instance.Raid and "R") or ((not instance.Raid) and "D")) .. tostring(diff)
	local prefs = SavedInstances.db.Indicators
	if not strfind(prefs[setting.."Text"], "ICON", 1, true) then
		return ColorCodeOpen(prefs[setting.."Color"]) .. prefs[setting.."Text"] .. FONTEND
	end
	local iconstring
	if prefs[setting.."Indicator"] == "BLANK" then
		iconstring = ""
	else
		iconstring = FONTEND .. Indicators[prefs[setting.."Indicator"]] .. ColorCodeOpen(prefs[setting.."Color"])
	end
	return ColorCodeOpen(prefs[setting.."Color"]) .. gsub(prefs[setting.."Text"], "ICON", iconstring) .. FONTEND
end

local function CullStaleInstances()
	for instance, a in pairs(SavedInstances.db.Instances) do
		a.Raid = a.Raid or false
		a.Expansion = a.Expansion or 0
		a.Order = a.Order or 1
		a.Show = a.Show or SavedInstances.db.Tooltip.NewInstanceShow
		for toon, b in pairs(a) do
			if type(b) == "table" then
				for difficulty, c in pairs(b) do
					if c.Expires < time() then
						SavedInstances.db.Instances[instance][toon][difficulty] = nil
					end
				end
				if TableLen(b) == 0 then
					SavedInstances.db.Instances[instance][toon] = nil
				end
			end
		end
	end
end

local function ShowIndicatorTooltip(cell, arg, ...)
	local instance = arg[1]
	local toon = arg[2]
	local diff = arg[3]
	if not instance or not toon or not diff then return end
	indicatortip = QTip:Acquire("SavedInstancesIndicatorTooltip", 2, "LEFT", "RIGHT")
	indicatortip:Clear()
	indicatortip:SetHeaderFont(tooltip:GetHeaderFont())
	local thisinstance = SavedInstances.db.Instances[instance]
	local nameline, _ = indicatortip:AddHeader()
	indicatortip:SetCell(nameline, 1, DifficultyString(instance, diff) .. " " .. GOLDFONT .. instance .. FONTEND, indicatortip:GetHeaderFont(), "LEFT", 2)
	indicatortip:AddHeader(ClassColorise(SavedInstances.db.Toons[toon].Class, strsplit(' ', toon)), thisinstance[toon][diff].ID)
	local categoryline, _ = indicatortip:AddLine()
	indicatortip:SetCell(categoryline, 1, YELLOWFONT .. SavedInstances.db.Categories[InstanceCategory(instance)].Name .. FONTEND, "LEFT", 2)
	indicatortip:AddLine(YELLOWFONT .. L["Time Left"] .. ":" .. FONTEND, SecondsToTime(thisinstance[toon][diff].Expires - time()))
	indicatortip:SetAutoHideDelay(0.1, tooltip)
	indicatortip:SmartAnchorTo(tooltip)
	indicatortip:Show()
end

local function ShowTooltip(anchorframe)
	if tooltip and tooltip:IsShown() then return end
	tooltip = QTip:Acquire("SavedInstancesTooltip", 1, "LEFT")
	tooltip:Clear()
	local hFont = tooltip:GetHeaderFont()
	local hFontPath, hFontSize
	hFontPath, hFontSize, _ = hFont:GetFont()
	hFont:SetFont(hFontPath, hFontSize, "OUTLINE")
	tooltip:SetHeaderFont(hFont)
	CullStaleInstances()
	if columns then columns = nil end
	local columns = { }
	if not SavedInstances.db.Toons[thisToon].Hide and SavedInstances.db.Tooltip.FreshToons then
		for diff = 1, 4 do
			columns[thisToon..diff] = columns[thisToon..diff] or tooltip:AddColumn("CENTER")
		end
	end
	local headLine, headCol = tooltip:AddHeader(GOLDFONT .. "SavedInstances" .. FONTEND)
	-- determining how many instances will be displayed per category
	local categorysize = { } -- remember the number of instances to be shown for each category
	local instancesaved = { } -- remember if each instance has been saved or not (boolean)
	local doseparator = false -- use this to determine whether to insert spaces or not
	for _, category in ipairs(OrderedCategories()) do
		categorysize[category] = 0
		for _, instance in ipairs(OrderedInstances(category)) do
			for toon, t in pairs(SavedInstances.db.Toons) do
				for diff = 1, 4 do
					if SavedInstances.db.Instances[instance][toon] and SavedInstances.db.Instances[instance][toon][diff] then
						instancesaved[instance] = true
						categorysize[category] = categorysize[category] + 1
					elseif SavedInstances.db.Tooltip.AllRaids and SavedInstances.db.Instances[instance].Show then
						categorysize[category] = categorysize[category] + 1
					end
				end
			end
		end
	end
	local categories = 0
	-- determining how many categories have instances that will be shown
	if SavedInstances.db.Tooltip.ShowCategories then
		for category, size in pairs(categorysize) do
			if size > 0 then
				categories = categories + 1
			end
		end
	end
	-- allocating tooltip space for instances, categories, and space between categories
	local categoryrow = { } -- remember where each category heading goes
	local instancerow = { } -- remember where each instance goes
	local firstcategory = true -- use this to skip spacing before the first category
	for _, category in ipairs(OrderedCategories()) do
		if categorysize[category] > 0 then
			if not firstcategory then
				tooltip:AddSeparator(6,0,0,0,0)
			end
			if (categories > 1 or SavedInstances.db.Tooltip.ShowSoloCategory) and (categorysize[category] > 0) then
				categoryrow[category], _ = tooltip:AddLine()

			end
			for _, instance in ipairs(OrderedInstances(category)) do
				for toon, t in pairs(SavedInstances.db.Toons) do
					for diff = 1, 4 do
						if SavedInstances.db.Instances[instance][toon] and SavedInstances.db.Instances[instance][toon][diff] then
							instancerow[instance] = instancerow[instance] or tooltip:AddLine()
							for diff = 1, 4 do
								columns[toon..diff] = columns[toon..diff] or tooltip:AddColumn("CENTER")
							end
						elseif SavedInstances.db.Tooltip.AllRaids and SavedInstances.db.Instances[instance].Show then
							instancerow[instance] = instancerow[instance] or tooltip:AddLine()
						end
					end
				end
			end
			firstcategory = false
		end
	end
	-- now printing instance data
	for instance, row in pairs(instancerow) do
		if instancesaved[instance] then
			tooltip:SetCell(instancerow[instance], 1, GOLDFONT .. instance .. FONTEND)
			for toon, t in pairs(SavedInstances.db.Toons) do
				if SavedInstances.db.Instances[instance][toon] then
					for diff = 1, 4 do
						local save = SavedInstances.db.Instances[instance][toon][diff]
						if save ~= nil then
							tooltip:SetCell(instancerow[instance], columns[toon..diff], DifficultyString(instance, diff))
							tooltip:SetCellScript(instancerow[instance], columns[toon..diff], "OnEnter", ShowIndicatorTooltip, {instance, toon, diff})
						end
					end
				end
			end
		elseif (not instancesaved[instance]) and (SavedInstances.db.Instances[instance].Show) and (SavedInstances.db.Tooltip.AllRaids) then
			tooltip:SetCell(instancerow[instance], 1, GRAYFONT .. instance .. FONTEND)
		end
	end
	-- toon names
	if TableLen(instancerow) == 0 then
		local noneLine = tooltip:AddLine()
		tooltip:SetCell(noneLine, 1, GRAYFONT .. NO_RAID_INSTANCES_SAVED .. FONTEND, "LEFT", tooltip:GetColumnCount())
	end
	if SavedInstances.db.Tooltip.FreshToons then
		for toon, t in pairs(SavedInstances.db.Toons) do
			if not SavedInstances.db.Toons[toon].Hide then
				for diff = 1, 4 do
					columns[toon..diff] = columns[toon..diff] or tooltip:AddColumn("CENTER")
				end
			end
		end
	end
	for toondiff, col in pairs(columns) do
		local toon = strsub(toondiff, 1, #toondiff-1)
		local diff = strsub(toondiff, #toondiff, #toondiff)
		if diff == "1" then
			tooltip:SetCell(headLine, col, ClassColorise(SavedInstances.db.Toons[toon].Class, select(1, strsplit(" - ", toon))), tooltip:GetHeaderFont(), "CENTER", 4)
		end
	end 
	-- we now know enough to put in the category names where necessary
	if SavedInstances.db.Tooltip.ShowCategories then
		for category, row in pairs(categoryrow) do
			if (categories > 1 or SavedInstances.db.Tooltip.ShowSoloCategory) and (categorysize[category] > 0) then
				tooltip:SetCell(categoryrow[category], 1, YELLOWFONT .. SavedInstances.db.Categories[category].Name .. FONTEND, "LEFT", tooltip:GetColumnCount())
			end
		end
	end
	-- finishing up, with hints
	if SavedInstances.db.Tooltip.ShowHints then
		tooltip:AddSeparator(8,0,0,0,0)
		local hintLine, hintCol
		hintLine, hintCol = tooltip:AddLine()
		tooltip:SetCell(hintLine, hintCol, L["|cffffff00Left-click|r to show Blizzard's Raid Information"], "LEFT", tooltip:GetColumnCount())
		hintLine, hintCol = tooltip:AddLine()
		tooltip:SetCell(hintLine, hintCol, L["|cffffff00Right-click|r to configure SavedInstances"], "LEFT", tooltip:GetColumnCount())
		hintLine, hintCol = tooltip:AddLine()
		tooltip:SetCell(hintLine, hintCol, L["Hover mouse on indicator for details"], "LEFT", tooltip:GetColumnCount())
	end
	tooltip:SetAutoHideDelay(0.1, anchorframe)
	tooltip:SmartAnchorTo(anchorframe)
	tooltip:Show()
end

-- global addon code below

function SavedInstances:OnInitialize()
	SavedInstancesDB = SavedInstancesDB or defaultDB
	-- begin backwards compatibility
	if not SavedInstancesDB.DBVersion then
		SavedInstancesDB = defaultDB
	end
	if SavedInstancesDB.DBVersion == 4 then
		SavedInstancesDB.DBVersion = 5
		SavedInstancesDB.Indicators.UnsavedIndicator = nil
		SavedInstancesDB.Indicators.UnsavedText = nil
		SavedInstancesDB.Indicators.UnsavedColor = nil
		for toon, a in pairs(SavedInstancesDB.Toons) do
			a.Class = strupper(gsub(a.Class, " ", ""))
			a.NumSaved = nil
		end
	end
	if SavedInstancesDB.DBVersion == 5 then
		SavedInstancesDB.DBVersion = 6
		SavedInstancesDB.Indicators = defaultDB.Indicators
		SavedInstancesDB.Categories = defaultDB.Categories
		SavedInstancesDB.Tooltip.ShowCategories = true
		SavedInstancesDB.Tooltip.ShowSoloCategory = false
		for instance, a in pairs(SavedInstancesDB.Instances) do
			for toon, b in pairs(SavedInstancesDB.Toons) do
				if a[toon] then
					if a[toon]["Normal"] then
						a[toon][1] = a[toon]["Normal"]
						a[toon]["Normal"] = nil
					end
					if a[toon]["Heroic"] then
						a[toon][2] = a[toon]["Heroic"]
						a[toon]["Heroic"] = nil
					end
				end
			end
		end
	end
	if SavedInstancesDB.DBVersion == 6 then
		SavedInstancesDB.DBVersion = 7
		SavedInstancesDB.Tooltip.ShowHints = true
	end
	if SavedInstancesDB.DBVersion ~= 7 then
		SavedInstancesDB = defaultDB
	end
	-- end backwards compatibilty
	self.db = db or SavedInstancesDB
	self.defaultDB = defaultDB
	self.db.Toons[thisToon] = self.db.Toons[thisToon] or { }
	self.db.Toons[thisToon].Class = select(2, UnitClass("player"))
	self.db.Toons[thisToon].Hide = self.db.Toons[thisToon].Hide or false
	self:SetupOptions()
	if LDB then
		self.launcher = LDB:NewDataObject("SavedInstances", {
			label = "SavedInstances",
			type = "launcher",
			icon = "Interface\\Addons\\SavedInstances\\icon.tga",
			OnClick = function(frame, button)
				if button == "LeftButton" then
					ToggleFriendsFrame(5) -- open Blizzard Raid window
					RaidInfoFrame:Show()
				else
					SavedInstances:ShowConfig()
				end
			end,
			OnEnter = function(self)
				ShowTooltip(self)
			end,
		})
	end
	if LDBIcon then
		LDBIcon:Register("SavedInstances", self.launcher, self.db.MinimapIcon)
	end
	self:RequestRefresh()
end

function SavedInstances:OnEnable()
	self:RegisterEvent("UPDATE_INSTANCE_INFO", "Refresh")
	self:RegisterEvent("RAID_INSTANCE_WELCOME", "RequestRefresh")
	self:RegisterEvent("CHAT_MSG_SYSTEM", "CheckSystemMessage")
end

function SavedInstances:OnDisable()
	self:UnregisterEvent("UPDATE_INSTANCE_INFO")
	self:UnregisterEvent("RAID_INSTANCE_WELCOME")
	self:UnregisterEvent("CHAT_MSG_SYSTEM")
end

function SavedInstances:CheckSystemMessage()
	if tostring(arg1) == INSTANCE_SAVED then
		self:RequestRefresh()
	end
end

function SavedInstances:RequestRefresh()
	RequestRaidInfo()
end

function SavedInstances:Refresh()
	CullStaleInstances()
	local numsaved = GetNumSavedInstances()
	if numsaved > 0 then
		for i = 1, numsaved do
			local name, id, expires, diff, locked, extended, mostsig, raid, players, diffname = GetSavedInstanceInfo(i)
			SavedInstances.db.Instances[name] = SavedInstances.db.Instances[name] or { }
			local instance = SavedInstances.db.Instances[name]
			instance.Raid = instance.Raid or raid
			instance.Expansion = instance.Expansion or 0
			instance.Order = instance.Order or 1
			instance.Show = instance.Show or SavedInstances.db.Tooltip.NewInstanceShow
			if locked then
				instance[thisToon] = instance[thisToon] or { }
				instance[thisToon][diff] = {
					ID = id,
					Expires = expires + time()
				}
			end
		end
	end
end

-- options functions below

local function IndicatorIconOptions(nextorder)
	return { order = nextorder, type = "select", width = "half", name = EMBLEM_SYMBOL, values = Indicators }
end

local function IndicatorTextOptions(nextorder)
	return { order = nextorder, type = "input", name = L["Text"], multiline = false }
end

local function IndicatorColorOptions(nextorder)
	return { order = nextorder, type = "color", width = "half", hasAlpha = false, name = COLOR,
		get = function(info)
			local r = SavedInstances.db.Indicators[info[#info]][1]
			local g = SavedInstances.db.Indicators[info[#info]][2]
			local b = SavedInstances.db.Indicators[info[#info]][3]
			return r, g, b, nil
		end,
		set = function(info, r, g, b, ...)
			SavedInstances.db.Indicators[info[#info]][1] = r
			SavedInstances.db.Indicators[info[#info]][2] = g
			SavedInstances.db.Indicators[info[#info]][3] = b
		end,
	}
end
					
-- options table below

SavedInstances.Options = {
	type = "group",
	name = "SavedInstances",
	handler = SavedInstances,
	args = {
		General = {
			order = 1,
			type = "group",
			name = L["General settings"],
			get = function(info)
					return SavedInstances.db.Tooltip[info[#info]]
			end,
			set = function(info, value)
					SavedInstances.db.Tooltip[info[#info]] = value
			end,
			args = {
				intro = {
					order = 1,
					type = "description",
					name = L["Track the instance IDs saved against your characters"],
				},
				MinimapIcon = {
					type = "toggle",
					name = L["Show minimap button"],
					desc = L["Show the SavedInstances minimap button"],
					order = 2,
					hidden = function() return not LDBIcon end,
					get = function(info) return not SavedInstances.db.MinimapIcon.hide end,
					set = function(info, value)
						SavedInstances.db.MinimapIcon.hide = not value
						if value then LDBIcon:Show("SavedInstances") else LDBIcon:Hide("SavedInstances") end
					end,
				},
				ShowHints = {
					type = "toggle",
					name = L["Show tooltip hints"],
					order = 3,
				},
				
				CategoriesHeader = {
					order = 4, 
					type = "header",
					name = L["Categories"],
				},
				ShowCategories = {
					type = "toggle",
					name = L["Show category names"],
					desc = L["Show category names in the tooltip"],
					order = 5,
				},
				ShowSoloCategory = {
					type = "toggle",
					name = L["Single category name"],
					desc = L["Show name for a category when all displayed instances belong only to that category"],
					order = 6,
					disabled = function()
						return not SavedInstances.db.Tooltip.ShowCategories
					end,
				},
				
				InstancesHeader = {
					order = 7, 
					type = "header",
					name = L["Instances"],
				},
				AllRaids = {
					type = "toggle",
					order = 8,
					name = L["Show all instances"],
					desc = L["Always show all instances, even if no characters are saved to them"],
				},
				NewInstanceShow = {
					type = "toggle",
					order = 9,
					name = L["Always show new instances"],
					desc = L["New instances are shown by default, even if no characters are saved to them"],
				},
				InstanceList = {
					order = 10,
					type = "multiselect",
					dialogControl = "Dropdown",
					width = "double",
					name = L["Show these instances when all instances are shown"],
					disabled = function() return not SavedInstances.db.Tooltip.AllRaids end,
					get = function(info, instance)
						return SavedInstances.db.Instances[instance].Show
					end,
					set = function(info, instance, value)
						SavedInstances.db.Instances[instance].Show = value
					end,
					values = function()
						local list = { }
						for instance, v in pairs(SavedInstances.db.Instances) do
							list[instance] = instance
						end
						return list
					end,
				},
				
				ToonHeader = {
					order = 11, 
					type = "header",
					name = L["Characters"],
				},
				FreshToons = {
					type = "toggle",
					name = L["Show non-saved characters"],
					desc = L["Show characters that have not been saved to any instances"],
					order = 12,
				},
				ToonList = {
					order = 13,
					type = "multiselect",
					dialogControl = "Dropdown",
					width = "double",
					name = L["Always hide these characters"],
					get = function(info, toon)
						return SavedInstances.db.Toons[toon].Hide
					end,
					set = function(info, toon, value)
						SavedInstances.db.Toons[toon].Hide = value
					end,
					values = function()
						local list = { }
						for k, v in pairs(SavedInstances.db.Toons) do
							list[k] = k
						end
						return list
					end,
				},
			},
		},
		Indicators = {
			order = 2,
			type = "group",
			name = L["Indicators"],
			get = function(info)
					return SavedInstances.db.Indicators[info[#info]]
			end,
			set = function(info, value)
					SavedInstances.db.Indicators[info[#info]] = value
			end,
			args = {
				Instructions = {
					order = 1,
					type = "description",
					name = L["You can combine icons and text in a single indicator if you wish. Simply choose an icon, and insert the word ICON into the text field. Anywhere the word ICON is found, the icon you chose will be substituted in."],
				},

				D1Header = {
					order = 2, 
					type = "header",
					name = DUNGEON_DIFFICULTY1,
				},
				D1Indicator = IndicatorIconOptions(3),
				D1Text = IndicatorTextOptions(4),
				D1Color = IndicatorColorOptions(5),

				D2Header = {
					order = 6, 
					type = "header",
					name = DUNGEON_DIFFICULTY2,
				},
				D2Indicator = IndicatorIconOptions(7),
				D2Text = IndicatorTextOptions(8),
				D2Color = IndicatorColorOptions(9),
				
				R1Header = {
					order = 10, 
					type = "header",
					name = RAID_DIFFICULTY1,
				},
				R1Indicator = IndicatorIconOptions(11),
				R1Text = IndicatorTextOptions(12),
				R1Color = IndicatorColorOptions(13),

				R2Header = {
					order = 14, 
					type = "header",
					name = RAID_DIFFICULTY2,
				},
				R2Indicator = IndicatorIconOptions(15),
				R2Text = IndicatorTextOptions(16),
				R2Color = IndicatorColorOptions(17),
				
				R3Header = {
					order = 18, 
					type = "header",
					name = RAID_DIFFICULTY3,
				},
				R3Indicator = IndicatorIconOptions(19),
				R3Text = IndicatorTextOptions(20),
				R3Color = IndicatorColorOptions(21),

				R4Header = {
					order = 22, 
					type = "header",
					name = RAID_DIFFICULTY4,
				},
				R4Indicator = IndicatorIconOptions(23),
				R4Text = IndicatorTextOptions(24),
				R4Color = IndicatorColorOptions(25),

				DefaultsHeader = {
					order = -2, 
					type = "header",
					name = DEFAULTS,
				},
				ResetButton = {
					order = -1,
					type = "execute",
					name = RESET_TO_DEFAULT,
					func = function()
						SavedInstances.db.Indicators = SavedInstances.defaultDB.Indicators
					end,
				},
			},
		},
		Categories = {
			order = 3,
			type = "group",
			name = L["Categories"],
			get = function(info)
					return SavedInstances.db.Categories[info[#info]]
			end,
			set = function(info, value)
					SavedInstances.db.Categories[info[#info]] = value
			end,
			args = {
				SelectedCategory = {
					order = 1,
					type = "select",
					width = "double",
					name = L["Selected category"],
					get = function(info, category)
						if not SavedInstances.selectedCategory then return end
						return SavedInstances.db.Categories[SavedInstances.selectedCategory].Order
					end,
					set = function(info, index)
						local categories = OrderedCategories()
						SavedInstances.selectedCategory = categories[index]
						SavedInstances.selectedInstance = nil
					end,
					values = function()
						local table = { }
						for i, v in ipairs(OrderedCategories()) do
							table[i] = i .. ". " .. SavedInstances.db.Categories[v].Name
						end
						return table
					end,
				},
				
				ReorderHeader = {
					order = 2, 
					type = "header",
					name = L["Category order"],
				},
				Position = {
					order = 3,
					type = "select",
					width = "double",
					name = L["Position in list"],
					disabled = function()
						return (SavedInstances.selectedCategory == nil)
					end,
					get = function(info, order)
						if SavedInstances.selectedCategory == nil then return end
						return SavedInstances.db.Categories[SavedInstances.selectedCategory].Order
					end,
					set = function(info, order)
						ReorderCategory(SavedInstances.db.Categories[SavedInstances.selectedCategory].Order, order)
					end,
					values = function()
						local table = { }
						for i, category in ipairs(OrderedCategories()) do
							table[i] = i .. ". " .. SavedInstances.db.Categories[category].Name
						end
						return table
					end,
				},

				DefaultsHeader = {
					order = -2, 
					type = "header",
					name = DEFAULTS,
				},
				ResetButton = {
					order = -1,
					type = "execute",
					name = RESET_TO_DEFAULT,
					func = function()
						SavedInstances.db.Categories = SavedInstances.defaultDB.Categories
					end,
				},
			},
		},
		Instances = {
			order = 4,
			type = "group",
			name = L["Instance options"],
			disabled = function()
				return (SavedInstances.selectedInstance == nil) or (InstanceCategory(SavedInstances.selectedInstance) ~= SavedInstances.selectedCategory)
			end,
			args = {
				SelectedCategory = {
					order = 1,
					type = "select",
					width = "double",
					name = L["Selected category"],
					disabled = false,
					get = function(info, category)
						if not SavedInstances.selectedCategory then return end
						return SavedInstances.db.Categories[SavedInstances.selectedCategory].Order
					end,
					set = function(info, index)
						local categories = OrderedCategories()
						SavedInstances.selectedCategory = categories[index]
						SavedInstances.selectedInstance = nil
					end,
					values = function()
						local table = { }
						for i, v in ipairs(OrderedCategories()) do
							table[i] = i .. ". " .. SavedInstances.db.Categories[v].Name
						end
						return table
					end,
				},
				SelectedInstance = {
					order = 2,
					type = "select",
					width = "double",
					name = L["Selected instance"],
					disabled = function()
						return (SavedInstances.selectedCategory == nil) or
							(TableLen(OrderedInstances(SavedInstances.selectedCategory)) == 0)
					end,
					get = function(info, order)
						if SavedInstances.selectedInstance == nil then return end
						return SavedInstances.db.Instances[SavedInstances.selectedInstance].Order
					end,
					set = function(info, index)
						for instance, v in pairs(SavedInstances.db.Instances) do
							if (InstanceCategory(instance) == SavedInstances.selectedCategory) and (v.Order == index) then
								SavedInstances.selectedInstance = instance
								break
							end
						end
					end,
					values = function()
						if SavedInstances.selectedCategory == nil then return { } end
						local table = { }
						for i, name in ipairs(OrderedInstances(SavedInstances.selectedCategory)) do
							table[i] = i .. ". " .. name
						end
						return table
					end,
				},

				OrderHeader = {
					order = 3, 
					type = "header",
					name = L["Instance order"],
				},
				Position = {
					order = 4,
					type = "select",
					width = "double",
					name = L["Position in list"],
					disabled = function()
						return (SavedInstances.selectedCategory == nil) or (SavedInstances.selectedInstance == nil)
					end,
					get = function(info, order)
						if SavedInstances.selectedInstance == nil then return end
						return SavedInstances.db.Instances[SavedInstances.selectedInstance].Order
					end,
					set = function(info, index)
						ReorderInstance(InstanceCategory(SavedInstances.selectedInstance), SavedInstances.db.Instances[SavedInstances.selectedInstance].Order, index)
					end,
					values = function()
						if (SavedInstances.selectedCategory == nil) or (SavedInstances.selectedInstance == nil) then return { } end
						local table = { }
						for i, instance in ipairs(OrderedInstances(SavedInstances.selectedCategory)) do
							table[i] = i .. ". " .. instance
						end
						return table
					end,
				},

				DetailsHeader = {
					order = 5, 
					type = "header",
					name = L["Instance details"],
				},
				Expansion = {
					order = 6,
					type = "select",
					name = L["Expansion"],
					get = function(info)
						if SavedInstances.selectedInstance then
							return SavedInstances.db.Instances[SavedInstances.selectedInstance].Expansion
						end
					end,
					set = function(info, expansion)
						local instance = SavedInstances.db.Instances[SavedInstances.selectedInstance]
						instance.Expansion = expansion
						instance.Raid = instance.Raid or false
						SavedInstances.selectedCategory = InstanceCategory(SavedInstances.selectedInstance)
					end,
					values = {
						[0] = EXPANSION_NAME0,
						[1] = EXPANSION_NAME1,
						[2] = EXPANSION_NAME2,
					},
				},
				Raid = {
					order = 7,
					type = "toggle",
					width = "half",
					name = LFG_TYPE_RAID,
					get = function(info)
						if SavedInstances.selectedInstance then
							return SavedInstances.db.Instances[SavedInstances.selectedInstance].Raid
						end
					end,
					set = function(info, value)
						SavedInstances.db.Instances[SavedInstances.selectedInstance].Raid = value
						SavedInstances.selectedCategory = InstanceCategory(SavedInstances.selectedInstance)
					end,
				},
				DeleteButton = {
					order = 8,
					type = "execute",
					width = "half",
					name = DELETE,
					func = function()
						local instance = SavedInstances.selectedInstance
						SavedInstances.selectedInstance = nil
						SavedInstances.db.Instances[instance] = nil
					end,
				},
			},
		},
	},
}

local optionsFrames = { }
function SavedInstances:SetupOptions()
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("SavedInstances", SavedInstances.Options)
	optionsFrames["General"] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SavedInstances", nil, nil, "General")
	optionsFrames["Indicators"] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SavedInstances", L["Indicators"], "SavedInstances", "Indicators")
	optionsFrames["Categories"] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SavedInstances", L["Categories"], "SavedInstances", "Categories")
	optionsFrames["Instances"] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SavedInstances", L["Instances"], "SavedInstances", "Instances")
end

function SavedInstances:ShowConfig()
	InterfaceOptionsFrame_OpenToCategory(optionsFrames["Instances"])
	InterfaceOptionsFrame_OpenToCategory(optionsFrames["General"])
end

